%{

/*
 Parser for PL/LOLCODE
 $Id: gram.y,v 1.14 2007/10/13 23:47:49 eggyknap Exp $
*/

#define YYPARSE_PARAM result
#include "pllolcode.h"
#include <stdio.h>

int  pllolcode_yylex(void);
void pllolcode_yyerror(char *s);
int  pllolcode_yyparse(void *result);

void executeProgram(lolNodeList *prg);
void executeList(lolNodeList *list);
void executeNode(lolNode *node);
/*nodeStruct makeNode(NodeType typ, nodeStruct child1, nodeStruct child2, NodeVal nodeVal);
void printNodeTree(nodeStruct prg); */
lolNode *lolMakeNode(NodeType typ, NodeVal nodeVal, lolNodeList *list);
void printNodeList(lolNodeList *list);
void printNode(lolNode *cur);

void yarnNode(lolNode *node);
void numbrNode(lolNode *node);
void numbarNode(lolNode *node);
void troofNode(lolNode *node);
void visibleNode(lolNode *node);
void foundyrNode(lolNode *node);

NodeVal tmpnval;
lolNodeList *curList;
%}

/* BISON Declarations */ %name-prefix="pllolcode_yy"

%union  {
	int numbrVal;
	float numbarVal;
	char *yarnVal;
	struct lolNodeList *nodeList;
	struct lolNode *node;
}

%type <nodeList> lol_program
%type <nodeList> lol_cmdblock
%type <node> lol_identifier
%type <node> lol_cmd
%type <yarnVal> YARN
%type <yarnVal> IDENTIFIER
%type <numbrVal> NUMBR
%type <numbarVal> NUMBAR
%type <numbarVal> VISDEBUG VISLOG VISNOTICE VISINFO VISWARNING VISEXCEPTION lol_visibleLevel

%token HAI KTHXBYE
%token YARN IDENTIFIER NUMBR NUMBAR
%token TROOFWIN TROOFFAIL
%token VISIBLE VISDEBUG VISLOG VISNOTICE VISINFO VISWARNING VISEXCEPTION
%token FOUNDYR

%%
lol_program:
	HAI lol_cmdblock KTHXBYE		{ executeProgram($2); }
	;

lol_cmdblock:
	lol_cmdblock lol_cmd	{ $$ = lolListAppend($1, $2); }
	| lol_cmd 		{ $$ = lolMakeList($1); }
	;

lol_cmd:
	VISIBLE lol_cmd		{ 
					tmpnval.numbrVal = NOTICE;
					$$ = lolMakeNode(node_VISIBLE, tmpnval, lolMakeList($2));
				}
	| VISIBLE lol_visibleLevel lol_cmd
				{
					tmpnval.numbrVal = $2;
					$$ = lolMakeNode(node_VISIBLE, tmpnval, lolMakeList($3));
				}
	| FOUNDYR lol_cmd	{ $$ = lolMakeNode(node_FOUNDYR, tmpnval, lolMakeList($2)); }
	| lol_identifier	{ $$ = $1; }
	;

lol_identifier:
	YARN			{
					tmpnval.yarnVal = pstrdup($1);
					$$ = lolMakeNode(node_YARN, tmpnval, NULL);
				}
	| NUMBAR		{ tmpnval.numbarVal = $1; $$ = lolMakeNode(node_NUMBAR, tmpnval, NULL); }
	| NUMBR			{ tmpnval.numbrVal = $1; $$ = lolMakeNode(node_NUMBR, tmpnval, NULL); }
	| TROOFWIN		{ tmpnval.numbrVal = 1; $$ = lolMakeNode(node_TROOF, tmpnval, NULL); }
	| TROOFFAIL		{ tmpnval.numbrVal = 0; $$ = lolMakeNode(node_TROOF, tmpnval, NULL); }
	| IDENTIFIER		{
					tmpnval.yarnVal = pstrdup(pllolcode_yytext);
					$$ = lolMakeNode(node_IDENT, tmpnval, NULL);
				}
	;

lol_visibleLevel:
	VISDEBUG		{ $$ = DEBUG1; }
	| VISLOG		{ $$ = LOG; }
	| VISNOTICE		{ $$ = NOTICE; }
	| VISINFO		{ $$ = INFO; }
	| VISWARNING		{ $$ = WARNING; }
	| VISEXCEPTION		{ $$ = ERROR; }
	;
%%

void
pllolcode_yyerror(char *s)
{
	    elog(ERROR, s);
}

/*extern int pllolcode_yy_flex_debug; */

void
executeList(lolNodeList *list)
{
	lolNode *cur;

	if (list == NULL) return;
	cur = list->head;
	while (cur != NULL) {
		executeNode(cur);
		cur = cur->next;
	}
}

void
executeProgram(lolNodeList *prg)
{
	elog(DEBUG3, "Executing program");
	/*printNodeList(prg); */
	if (prg == NULL)  {
		elog(INFO, "NULL parse tree");
		return;
	}
	executeList(prg);
}

void
visibleNode(lolNode *node)
{
	executeList(node->list);
	elog(node->nodeVal.numbrVal, lolVarGetString(lolGetVar("IT")));
}

void
foundyrNode(lolNode *node)
{
	executeList(node->list);
	returnVal = lolGetVar("IT");
}

void
yarnNode(lolNode *node)
{
	lolSetVar("IT", ident_YARN, node->nodeVal);
}

void
numbrNode(lolNode *node) 
{
	lolSetVar("IT", ident_NUMBR, node->nodeVal);
}

void
numbarNode(lolNode *node)
{
	lolSetVar("IT", ident_NUMBAR, node->nodeVal);
}

void
troofNode(lolNode *node)
{
	lolSetVar("IT", ident_TROOF, node->nodeVal);
}

void
identNode(lolNode *node)
{
	lolIdent *cur;
	NodeVal val;

	cur = lolGetVar(node->nodeVal.yarnVal);
	if (cur != NULL)
		lolSetVar("IT", cur->type, cur->value);
	else 
	{
		val.numbrVal = 0;
		lolSetVar("IT", ident_NUMBR, val);
	}
		
}

void
executeNode(lolNode *node)
{
	switch (node->node_type) {
		case node_VISIBLE:
			visibleNode(node);
			break;
		case node_YARN:
			yarnNode(node);
			break;
		case node_NUMBR:
			numbrNode(node);
			break;
		case node_NUMBAR:
			numbarNode(node);
			break;
		case node_TROOF:
			troofNode(node);
			break;
		case node_FOUNDYR:
			foundyrNode(node);
			break;
		case node_IDENT:
			identNode(node);
			break;	
		default:
			elog(INFO, "Cannot execute node type");
			printNode(node);
			break;
	}
}

lolNodeList *
lolListAppend(lolNodeList *list, lolNode *node)
{
	if (list == NULL) { return NULL; }
	node->next = NULL;
	if (list->tail == NULL) {
		list->head = node;
		list->tail = node;
		list->length = 1;
		return list;
	}
	list->tail->next = node;
	list->tail = node;
	list->length++;
	return list;
}

lolNodeList *
lolMakeList(lolNode *node)
{
	lolNodeList *list;
	
	list = palloc(sizeof(lolNodeList));
	list->head = node;
	list->tail = node;
	list->length = 1;
	return list;	
} 
lolNode *
lolMakeNode(NodeType typ, NodeVal nodeVal, lolNodeList *list)
{
	lolNode *node;

	node = palloc(sizeof(lolNode));
	node->node_type = typ;
	node->nodeVal = nodeVal;
	node->list = list;
	node->next = NULL;
	return node;
}
void
printNode(lolNode *cur)
{
	char tmpString[255];

	if (cur == NULL) {
		elog(INFO, "NULL node");
		return;
	}
	switch (cur->node_type) {
		case node_CMDBLOCK:
			elog(INFO, "node_CMDBLOCK");
			printNodeList(cur->list);
			break;
		case node_VISIBLE:
			elog(INFO, "node_VISIBLE");
			printNodeList(cur->list);
			break;
		case node_YARN:
			elog(INFO, "node_YARN");
			snprintf(tmpString, 255, "value: %s", cur->nodeVal.yarnVal);
			elog(INFO, tmpString);
			break;
		case node_NUMBR:
			elog(INFO, "node_NUMBR");
			snprintf(tmpString, 255, "value: %d", cur->nodeVal.numbrVal);
			elog(INFO, tmpString);
			break;
		case node_NUMBAR:
			elog(INFO, "node_NUMBAR");
			snprintf(tmpString, 255, "value: %f", cur->nodeVal.numbarVal);
			elog(INFO, tmpString);
			break;
		case node_FOUNDYR:
			elog(INFO, "node_FOUNDYR");
			printNodeList(cur->list);
			break;
		case node_IDENT:
			elog(INFO, "node_IDENT: name = %s", cur->nodeVal.yarnVal);
			break;	
		default:
			elog(INFO, "Unknown node type");
	}
}

void
printNodeList(lolNodeList *list)
{
	lolNode *cur;

	elog(INFO, "Start of list");
	if (list == NULL) 
	{
		elog(INFO, "NULL node list");
		return;
	}
	cur = list->head;
	if (cur == NULL) {
		elog(INFO, "NULL first node");
		return;
	}
	while (cur->next != NULL) {
		elog(INFO, "List Node");
		printNode(cur);
	}
	printNode(cur);
	elog(INFO, "End of list");
}
