/*
 * $Id: pllolcode.c,v 1.8 2007/10/13 13:32:15 eggyknap Exp $
 */

#include "pllolcode.h"
#include "executor/spi.h"
#include "commands/trigger.h"
#include "fmgr.h"
#include "access/heapam.h"
#include "utils/syscache.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_type.h"
#include "utils/lsyscache.h"
#include "utils/memutils.h"

PG_MODULE_MAGIC;

extern int pllolcode_yyparse(void *result);
lolIdent *returnVal;

Datum pl_lolcode_call_handler(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(pl_lolcode_call_handler);

Datum
pl_lolcode_call_handler(PG_FUNCTION_ARGS)
{
	/* TODO: return error if you try to use this as a trigger, for now */
	FmgrInfo flinfo, tmp_flinfo;

	Form_pg_type typeStruct;
	Form_pg_proc procStruct;
	HeapTuple typeTup;
	HeapTuple procTup;
	Oid resultTypeIOParam, returnTypeOID; 
	char typeName[NAMEDATALEN + 1];
	Datum procsrcdatum;
	char *proc_source;
	char arg_name[255];
	NodeVal nodeVal;
	bool isnull, isbool;
	Datum retval = 0;
	int i;

	returnVal = NULL;
	if (SPI_connect() != SPI_OK_CONNECT) 
		ereport(ERROR, (errcode(ERRCODE_CONNECTION_EXCEPTION),
			errmsg("Couldn't connect to SPI"))); 

	/* TODO: Generally reorganize this code -- it should do one thing, finish it, and then do another. Right now it's kinda all over */
	/* TODO: release these tuples earlier, because it seems like a good idea */
	/* Get function tuple */
	procTup = SearchSysCache(PROCOID, ObjectIdGetDatum(fcinfo->flinfo->fn_oid), 0, 0, 0);
	if (!HeapTupleIsValid(procTup)) elog(ERROR, "Cache lookup failed for function %u", fcinfo->flinfo->fn_oid);
	procStruct = (Form_pg_proc) GETSTRUCT(procTup);

	/* Get return type tuple */
	typeTup = SearchSysCache(TYPEOID, ObjectIdGetDatum(procStruct->prorettype), 0, 0, 0);
	if (!HeapTupleIsValid(typeTup)) elog(ERROR, "Cache lookup failed for type %u", procStruct->prorettype);
	typeStruct = (Form_pg_type) GETSTRUCT(typeTup);

	MemSet(typeName, 0, NAMEDATALEN + 1);
	strncpy(typeName, typeStruct->typname.data, NAMEDATALEN);
	elog(DEBUG3, "PL/LOLCODE function return type name: %s", typeName);

	resultTypeIOParam = getTypeIOParam(typeTup);

	fmgr_info_cxt(typeStruct->typinput, &flinfo, CurTransactionContext);

	ReleaseSysCache(typeTup);

	lolInitVars();

	/* Set a lolIdent variable for each input argument */
	for (i = 0; i < procStruct->pronargs; i++)
	{
		typeTup = SearchSysCache(TYPEOID, ObjectIdGetDatum(procStruct->proargtypes.values[i]), 0, 0, 0);
		if (!HeapTupleIsValid(typeTup))
			elog(ERROR,"Cache lookup failed for type %u", procStruct->proargtypes.values[i]);
		if (procStruct->proargtypes.values[i] == BOOLOID)
		isbool = true;
	else
		isbool = false;		
	typeStruct = (Form_pg_type) GETSTRUCT(typeTup);
	if (typeStruct->typtype == 'p' || typeStruct->typtype == 'c')
		elog(ERROR, "Cannot support pseudo-types or composite types as function arguments");
	fmgr_info_cxt(typeStruct->typoutput, &tmp_flinfo, CurTransactionContext);	
	ReleaseSysCache(typeTup);

	/* All arguments are named LOL1, LOL2, etc. */
	snprintf(arg_name, 255, "LOL%d", i+1);
	nodeVal.yarnVal = OutputFunctionCall(&tmp_flinfo, fcinfo->arg[i]), 't';
	if (isbool)
		{
			if (nodeVal.yarnVal[0] == 't')
				nodeVal.numbrVal = 1;
			else
				nodeVal.numbrVal = 0;
			lolSetVar(arg_name, ident_TROOF, nodeVal);		
		}
		else 
			lolSetVar(arg_name, ident_YARN, nodeVal);
	}

	procsrcdatum = SysCacheGetAttr(PROCOID, procTup, Anum_pg_proc_prosrc, &isnull);
	if (isnull)
		elog(ERROR, "Function source is null");

	proc_source = DatumGetCString(DirectFunctionCall1(textout, procsrcdatum));
	elog(DEBUG3, proc_source);

	returnTypeOID = procStruct->prorettype;

	ReleaseSysCache(procTup);

	pllolcode_yy_scan_string(proc_source);
	pllolcode_yyparse(NULL);
	pllolcode_yylex_destroy();

	if (returnTypeOID == BOOLOID)
		retval = InputFunctionCall(&flinfo, lolVarGetTroof(returnVal) == lolWIN ? "TRUE" : "FALSE", resultTypeIOParam, -1);
	else
		retval = InputFunctionCall(&flinfo, lolVarGetString(returnVal), resultTypeIOParam, -1);

	SPI_finish();

	return retval;
}
