/*
 * PL/LOLCODE Symbol table functions
 * $Id: pllolcode_symbol.c,v 1.3 2007/10/13 13:32:15 eggyknap Exp $
 */

#include "pllolcode.h"
#include "postgres.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define LOL_MAX_VARS 255

lolIdent *IT;
/* TODO : Change this to some other implementation without a limited number of variables */
lolIdent lolVars[LOL_MAX_VARS];
int lolVarCount;

void
printSymbolTable(void)
{
	int i;
	elog(INFO, "Count; %d", lolVarCount);
	for (i = 0; i < lolVarCount; i++)
	{
		elog(INFO, "Name: %s", lolVars[i].name);
		switch (lolVars[i].type) {
			case ident_YARN:
				elog(INFO, "Yarn, value %s", lolVars[i].value.yarnVal);
				break;
			case ident_NUMBR:
				elog(INFO, "Numbr, value %d", lolVars[i].value.numbrVal);
				break;
			case ident_NUMBAR:
				elog(INFO, "Numbar value %f", lolVars[i].value.numbarVal);
				break;
			case ident_TROOF:
				elog(INFO, "Troof, value %s", (lolVars[i].value.numbrVal == 1) ? "WIN" : "FAIL");
				break;
			default:
				elog(INFO, "Unknown type %d", lolVars[i].type);
		}
	}
}

void
lolInitVars()
{
	NodeVal val;

	val.numbrVal = 0;
	IT = &(lolVars[0]);
	IT->name = pstrdup("IT");
	IT->type = ident_NUMBR;
	IT->value = val;
	lolVarCount = 1;

/*	printSymbolTable(); */
}

lolIdent *
lolGetVar(char *name)
{
	int i;

	if (name == NULL) return NULL;
	for (i = 0; i < lolVarCount; i++)
	{
		if (strncmp(name, lolVars[i].name, 255) == 0)
			return &(lolVars[i]);
	}
}

void
lolSetVar(char *name, IdentType type, NodeVal value)
{
	int i;

	for(i = 0; i < lolVarCount; i++)
	{
		if (strncmp(name, lolVars[i].name, 255) == 0)
			break;
	}
	if (i >= lolVarCount)
	{
		lolVarCount++;
		if (lolVarCount >= 255)
			elog(ERROR, "Too many variables required for LOLCODE function");
		lolVars[i].name = pstrdup(name);
	}
	lolVars[i].type = type;
	lolVars[i].value = value;
	/*printSymbolTable(); */
}

/* Handle string type casting */
char *
lolVarGetString(lolIdent *var)
{
	char *tmpString;

	/* TODO: Handle strings of more than 255 chars */
	tmpString = palloc(255);
	if (var == NULL) return NULL;
	switch (var->type) {
		case ident_YARN:
			return var->value.yarnVal;
		case ident_NUMBR:
			snprintf(tmpString, 255, "%d", var->value.numbrVal);
			return tmpString;
		case ident_NUMBAR:
			snprintf(tmpString, 255, "%f", var->value.numbarVal);
			return tmpString;
		case ident_TROOF:
			if (var->value.numbrVal != 0)
				snprintf(tmpString, 255, "WIN");
			else
				snprintf(tmpString, 255, "FAIL");
			return tmpString;
	}
	return NULL;
}

/* TODO: Error checking */
/* Type cast to NUMBR */
int
lolVarGetNumbr(lolIdent *var) {
	if (var == NULL) return 0;
	switch (var->type) {
		case ident_NUMBR:
			return var->value.numbrVal;
		case ident_NUMBAR:
			return (int) var->value.numbarVal;
		case ident_TROOF:
			if (var->value.numbrVal != 0)
				return 1;
			else return 0;
		case ident_YARN:
			/* Do I have to cast long int to int? */
			return (int) strtol(var->value.yarnVal, NULL, 10);
	}
	return 0;
}

/* Type cast to NUMBAR */
float
lolVarGetNumbar(lolIdent *var) {
	if (var == NULL) return 0;
	switch (var->type) {
		case ident_NUMBAR:
			return var->value.numbarVal;
		case ident_NUMBR:
			return (float) var->value.numbrVal;
		case ident_TROOF:
			if (var->value.numbrVal != 0)
				return 1.0;
			else return 0.0;
		case ident_YARN:
			return strtof(var->value.yarnVal, NULL);
	}
	return 0.0;
}

/* Type cast to TROOF */
lolTroof
lolVarGetTroof(lolIdent *var) {
	if (var == NULL) return lolFAIL;
	switch (var->type) {
		case ident_TROOF:
		case ident_NUMBAR:
			return (var->value.numbarVal != 0) ? lolWIN : lolFAIL;
		case ident_NUMBR:
			return (var->value.numbrVal != 0) ? lolWIN : lolFAIL;
		case ident_YARN:
			/* TODO: Error checking */
			return (strlen(var->value.yarnVal) > 0) ? lolWIN : lolFAIL;
	}
	return lolFAIL;
}
