%{
#include "pllolcode_lang.h"
#include "pllolcode.h"
#include "postgres.h"
#include <string.h>
#include <stdlib.h>
#define LOL_LEX_DEBUG 1

#define YYSTACK_FREE pfree
#define YYSTACK_MALLOC palloc
#define YYFREE pfree
#define YYMALLOC palloc

void *yyalloc(yy_size_t len) { return palloc(len); }
void *yyrealloc(void *ptr, yy_size_t len) { return repalloc(ptr, len); }
void yyfree(void *ptr) { pfree(ptr); }

int  pllolcode_yylex(void);
void pllolcode_yyerror(char *);

/* These declarations get rid of compiler warnings */
FILE *pllolcode_yyget_in(void);
FILE *pllolcode_yyget_out(void);
int   pllolcode_yyget_leng(void);
char *pllolcode_yyget_text(void);
void  pllolcode_yyset_lineno(int);
int   pllolcode_yyget_lineno(void);
void  pllolcode_yyset_in(FILE *);
void  pllolcode_yyset_out(FILE *);
int   pllolcode_yyget_debug(void);
void  pllolcode_yyset_debug(int);
int   pllolcode_yylex_destroy(void);
void  lol_lex_debug(char *s);

void pllolcode_scanner_init(const char *str);
void pllolcode_scanner_finish(void);

%}

%option nodefault
%option nounput
%option prefix="pllolcode_yy"
%option noyyalloc noyyrealloc noyyfree
/* %option debug */

%x LOL_STRING
%x LOL_LINECOMMENT
%x LOL_BLOCKCMT

%%
	/* todo: make this more robust (e.g. buffer length checking) */
	char *stringptr;

\"			{ 
				/* TODO: Quit using stringbuf, use yytext. alloc a new char* for the contents and stick it in pllolcode_yyval.yarnVal */
				BEGIN LOL_STRING;
				memset(stringbuf, 0, 255);
				stringptr = stringbuf;
			}
<LOL_STRING>\\n		{ *stringptr++ = '\n'; }
<LOL_STRING>\\t		{ *stringptr++ = '\t'; }
<LOL_STRING>\\\"	{ *stringptr++ = '\"'; }
<LOL_STRING>\"		{ 
				*stringptr = 0;
				BEGIN 0;
				lol_lex_debug("yylex found string"); lol_lex_debug(stringbuf);
				pllolcode_yylval.yarnVal = stringbuf;
				return YARN;
			}

<LOL_STRING>\n		{ lol_lex_debug("invalid string -- newline before closing quotes"); exit(1); }
<LOL_STRING>.		{ *stringptr++ = *yytext; }

BTW			{ BEGIN LOL_LINECOMMENT; }
<LOL_LINECOMMENT>\n	{ BEGIN 0; lol_lex_debug("yylex: line comment"); }
<LOL_LINECOMMENT>.	{}

OBTW			{ BEGIN LOL_BLOCKCMT; }
<LOL_BLOCKCMT>TLDR	{ BEGIN 0; lol_lex_debug("yylex: block comment"); }
<LOL_BLOCKCMT>\n	{}
<LOL_BLOCKCMT>.		{}

HAI			{ lol_lex_debug("yylex: HAI"); return HAI; }
KTHXBYE			{ lol_lex_debug("yylex: KTHXBYE"); return KTHXBYE; }

VISIBLE			{ lol_lex_debug("yylex: VISIBLE"); return VISIBLE; }
DEBUG			{ return VISDEBUG; }
LOG			{ return VISLOG; }
NOTICE			{ return VISNOTICE; }
INFO			{ return VISINFO; }
WARNING			{ return VISWARNING; }
EXCEPTION		{ return VISEXCEPTION; }

WIN			{ return TROOFWIN; }
FAIL			{ return TROOFFAIL; }

"FOUND YR"		{ lol_lex_debug("yylex: FOUND YR"); return FOUNDYR; }

[A-Za-z][A-Za-z0-9_]+		{ lol_lex_debug("yylex: IDENTIFIER"); return IDENTIFIER; }
[\+-]?[0-9]+			{
					lol_lex_debug("yylex: NUMBR");
					pllolcode_yylval.numbrVal = atoi(yytext);
					return NUMBR;
				}
[\+-]?[0-9\.]+			{
					lol_lex_debug("yylex: NUMBAR");
					pllolcode_yylval.numbarVal = strtof(yytext, NULL);
					return NUMBAR;
				}

[ \t\n]+		;
.			{ pllolcode_yyerror("Parse error: Unrecognized character"); }
%%

/*IT			{ lol_lex_debug("yylex: IT"); return IT; }
"CAN HAS"		{ lol_lex_debug("yylex: CANHAS"); return CANHAS; }
"GIMMEH"		{ lol_lex_debug("yylex: GIMMEH"); return GIMMEH; }
"OUTTA"			{ lol_lex_debug("yylex: OUTTA"); return OUTTA; }
"DIAF"			{ lol_lex_debug("yylex: DIAF"); return DIAF; }
"BYES"			{ lol_lex_debug("yylex: BYES"); return BYES; }
KTHX			{ lol_lex_debug("yylex: KTHX"); return KTHX; }
"GTFO"			{ lol_lex_debug("yylex: GTFO"); return GTFO; }
"IM IN YR"		{ lol_lex_debug("yylex: IMINYR"); return IMINYR; }
"I HAS A"		{ lol_lex_debug("yylex: IHASA"); return IHASA; }
"IN MAH"		{ lol_lex_debug("yylex: INMAH"); return INMAH; }
"LOL"			{ lol_lex_debug("yylex: LOL"); return LOL; }
"R"			{ lol_lex_debug("yylex: R"); return R; }

"MEEBE"			{ lol_lex_debug("yylex: MEEBE"); return MEEBE; }
"O RLY?"		{ lol_lex_debug("yylex: O RLY?"); return ORLY; }
"OIC"			{ lol_lex_debug("yylex: OIC"); return OIC; }
"YA RLY"		{ lol_lex_debug("yylex: YA RLY"); return YARLY; }
"NO WAI"		{ lol_lex_debug("yylex: NO WAI"); return NOWAI; }
"MEBBE"			{ lol_lex_debug("yylex: MEBBE"); return MEBBE; }
"NOT"			{ lol_lex_debug("yylex: NOT"); return NOT; }
"BIGR THAN"		{ lol_lex_debug("yylex: BIGR"); return BIGR; }
"SMALR THAN"		{ lol_lex_debug("yylex: SMALR"); return SMALR; }
"BOTH SAEM"		{ lol_lex_debug("yylex: BOTH SAEM"); return BOTHSAEM; }
"DIFFRINT"		{ lol_lex_debug("yylex: DIFFRINT"); return DIFFRINT; }
"AN"			{ lol_lex_debug("yylex: AN"); return AN; }

"SUM OF"		{ lol_lex_debug("yylex: SUMOF"); return SUMOF; }
"DIFF OF"		{ lol_lex_debug("yylex: DIFFOF"); return DIFFOF; }
"PRODUKT OF"		{ lol_lex_debug("yylex: PRODUKTOF"); return PRODUKTOF; }
"QUOSHUNT OF"		{ lol_lex_debug("yylex: QUOSHUNTOF"); return QUOSHUNTOF; }
"MOD OF"		{ lol_lex_debug("yylex: MODOF"); return MODOF; }
"BIGGR OF"		{ lol_lex_debug("yylex: BIGGROF"); return BIGGROF; }
"SMALLR OF"		{ lol_lex_debug("yylex: SMALLROF"); return SMALLROF; }

"NOT BIGR THAN"		{ lol_lex_debug("yylex: NOTBIGR"); return NOTBIGR; }
"NOT SMALR THAN"	{ lol_lex_debug("yylex: NOTSMALR"); return NOTSMALR; }
"AND"			{ lol_lex_debug("yylex: AND"); return AND; }
"OR"			{ lol_lex_debug("yylex: OR"); return OR; }
"XOR"			{ lol_lex_debug("yylex: XOR"); return XOR; }
"UP"			{ lol_lex_debug("yylex: UP"); return UP; }
"UPZ"			{ lol_lex_debug("yylex: UPZ"); return UPZ; }
"!!"			{ lol_lex_debug("yylex: BANG"); return BANG; }
NOOB			{ lol_lex_debug("yylex: NOOB"); return NOOB; }
"NERF"			{ lol_lex_debug("yylex: NERF"); return NERF; }
"NERFZ"			{ lol_lex_debug("yylex: NERFZ"); return NERFZ; }
"TIEMZ"			{ lol_lex_debug("yylex: TIEMZ"); return TIEMZ; }
"TIEMZD"		{ lol_lex_debug("yylex: TIEMZD"); return TIEMZD; }
"OVAR"			{ lol_lex_debug("yylex: OVAR"); return OVAR; }
"OVARZ"			{ lol_lex_debug("yylex: OVARZ"); return OVARZ; }
"DATABUKKIT"		{ lol_lex_debug("yylex: DATABUKKIT"); return DATABUKKIT; }
"STDIN"			{ lol_lex_debug("yylex: STD_IN"); return STD_IN; }
"FOUND YR"		{ lol_lex_debug("yylex: FOUNDYR"); return FOUNDYR; }
\.\.\.[ \t]*\n		{ lol_lex_debug("yylex: Continuation"); }
[,\n]+			{ lol_lex_debug("yylex: CMDEND"); return CMDEND; }
\?			{ lol_lex_debug("yylex: ? "); return QM; }
*/

void lol_lex_debug(char *s) {
	if (LOL_LEX_DEBUG)
		elog(DEBUG5,s);
}

int pllolcode_yywrap(void) {
	return 1;
}
